#include "gibbsMedLDA-MT.h"
#include "stats.h"
#include "cokus.h"
#include "utils.h"
#include <math.h>
#include <assert.h>
#include <stdio.h>
#include <string.h>
#include <vector>
#include <map>
#include <time.h>
using namespace std;

void MultiTaskGibbsMedLDA::set_default_values()
{
	m_ptrnstats			= NULL;
	m_ptststats			= NULL;

	m_nK				= 10; // number of topics, K
	m_nV				= 0;
	m_nLabelNum			= 0; // number of classes, |C|
	m_dAlpha			= 1; 
	m_dBeta				= 1;
	m_dNuSqInv          = 1;
	m_dC				= 1;
	m_dB				= 0;
	m_nBurnIn           = 100;
	m_nMaxSamples       = 100;

	m_dEta				= NULL; 
	m_dLambdaInv        = NULL;
	m_dCov              = NULL;
	m_dCovInv           = NULL;
	m_dLowerTriangle    = NULL;
	m_dPr				= NULL; 
	m_mrg				= NULL;
	m_dPhi				= NULL;
	m_dTheta			= NULL;
	m_dEtabk_           = NULL;
	m_dDelta_			= NULL;
	m_dDiscValue_		= NULL;
	m_dSampleEta        = NULL;
	m_pIGsamplerG       = NULL;
}

MultiTaskGibbsMedLDA::~MultiTaskGibbsMedLDA()
{
	if (m_dEta) delete[] m_dEta;
	m_dEta = NULL;

	if ( m_dCov ) {
		for ( int i=0; i<m_nK; i++ ) {
			delete[] m_dCov[i];
			delete[] m_dCovInv[i];
			delete[] m_dLowerTriangle[i];
		}
		delete[] m_dCov;
		delete[] m_dCovInv;
		delete[] m_dLowerTriangle;
	}

	if ( m_dEtabk_ ) delete[] m_dEtabk_;
	if ( m_dDelta_ ) delete[] m_dDelta_;
	if ( m_dDiscValue_ ) delete[] m_dDiscValue_;
	if ( m_dSampleEta  ) {
		for ( int i=0; i<m_nMaxSamples; i++ ) {
			delete[] m_dSampleEta[i];
		}
		delete[] m_dSampleEta;
	}

	if (m_dPr) delete[] m_dPr;
	m_dPr = NULL;

	if (m_mrg) delete[] m_mrg;
	m_mrg = NULL;

	if (m_dPhi) {
		for(int i = 0; i < m_nK; ++i)
			delete[] m_dPhi[i];
		delete[] m_dPhi;
	}
	m_dPhi = NULL;

	if (m_dTheta) delete[] m_dTheta;
	m_dTheta = NULL;

	if (m_ptrnstats) delete m_ptrnstats;
	m_ptrnstats = NULL;

	if (m_ptststats) delete m_ptststats;
	m_ptststats	= NULL;

	if ( m_pMVGaussian ) delete m_pMVGaussian;
}

void MultiTaskGibbsMedLDA::init_est(int refLabel)
{
	Corpus* corpus = new Corpus();
	corpus->read_data_binary(m_params->train_filename, refLabel);
	init_est(corpus);
}

void MultiTaskGibbsMedLDA::init_est(Corpus* corpus)
{
	m_nK 			= m_params->NTOPICS;
	m_nLabelNum		= m_params->NLABELS,
	m_dC			= m_params->INITIAL_C;
	m_dDeltaEll     = m_params->DELTA_ELL;
	m_dNuSqInv      = m_params->INITIAL_NUSQINV; //1 / m_dC;
	m_dAlpha 		= m_params->INITIAL_ALPHA / m_params->NTOPICS;
	m_dBeta 		= m_params->INITIAL_BETA; 
	m_nBurnIn       = m_params->BURN_IN;
	m_nMaxSamples   = m_params->MAX_SAMPLES;

	m_ptrnstats = new Stats();
	m_ptrnstats->new_stats(corpus, m_params);
	m_nV = m_ptrnstats->nV;
	m_dVbeta = m_nV * m_dBeta;

	const int m_nD = m_ptrnstats->nD;
	int **m_nw		= m_ptrnstats->nw;
	int *m_nwsum	= m_ptrnstats->nwsum;
	int **m_nd		= m_ptrnstats->nd;

	if (!m_dEta) {
		m_dEta = new double*[m_nLabelNum];
		for ( int y=0; y<m_nLabelNum; y++ ) {
			m_dEta[y] = new double[m_nK];
			memset(m_dEta[y], 0, sizeof(double) * m_nK);
		}
	}
	if (!m_dCov) {
		m_dCov = new double*[m_nK];
		m_dCovInv = new double*[m_nK];
		m_dLowerTriangle = new double*[m_nK];
		for ( int k=0; k<m_nK; k++ ) {
			m_dCov[k] = new double[m_nK];
			m_dCovInv[k] = new double[m_nK];
			m_dLowerTriangle[k] = new double[m_nK];
			memset(m_dCov[k], 0, sizeof(double) * m_nK);
			memset(m_dCovInv[k], 0, sizeof(double) * m_nK);
			memset(m_dLowerTriangle[k], 0, sizeof(double) * m_nK);
		}
	}
	m_dSampleEta = new double**[m_nMaxSamples];
	for ( int i=0; i<m_nMaxSamples; i++ ) {
		m_dSampleEta[i] = new double*[m_nLabelNum];
		for ( int y=0; y<m_nLabelNum; y++ ) {
			m_dSampleEta[i][y] = new double[ m_nK ];
			memset(m_dSampleEta[i][y], 0, sizeof(double) * m_nK);
		}
	}

	if (!m_dLambdaInv) {
		m_dLambdaInv = new double*[m_nLabelNum];
		for ( int y=0; y<m_nLabelNum; y++ ) {
			m_dLambdaInv[y] = new double[corpus->num_docs];
			memset(m_dLambdaInv[y], 1, sizeof(double)*corpus->num_docs);
		}
	}
	if (!m_dEtabk_) m_dEtabk_ = new double[m_nK];
	memset(m_dEtabk_, 0, sizeof(double)*m_nK);

	if (!m_dPr) m_dPr = new double[m_nK];
	if (!m_mrg) m_mrg = new double[m_nK];

	int N, topic;
	int* zptr = NULL;
	for(int d = 0; d < m_nD; ++d) {
		Document& dc = corpus->docs[d];
		N = dc.length; 
		zptr = m_ptrnstats->z[d];
		for(int n = 0; n < N; ++n) {
			topic = (int)(myrand() * m_nK);	
			*zptr++ = topic;
			m_nw[topic][dc.words[n]] += 1;
			m_nwsum[topic] += 1;
			m_nd[d][topic] += 1;
		}
	}
	m_dTrainTime_Eta = 0;
	m_dTrainTime_Lambda = 0;
}

void MultiTaskGibbsMedLDA::init_estc(const char* prev_state_file)
{
	m_nK 			= m_params->NTOPICS; 
	m_nLabelNum		= m_params->NLABELS,
	m_dC			= m_params->INITIAL_C;
	m_dDeltaEll     = m_params->DELTA_ELL;
	m_dNuSqInv      = m_params->INITIAL_NUSQINV; //1 / m_dC;
	m_dAlpha 		= m_params->INITIAL_ALPHA / m_params->NTOPICS;
	m_dBeta 		= m_params->INITIAL_BETA; 
	m_dVbeta		= m_nV * m_dBeta;

	if (!m_ptrnstats) load_model(prev_state_file);

	if (!m_dPr) m_dPr = new double[m_nK];
	if (!m_mrg) m_mrg = new double[m_nK];

	m_dTrainTime_Eta = 0;
	m_dTrainTime_Lambda = 0;
}

void MultiTaskGibbsMedLDA::init_inf(Corpus* corpus)
{
	m_nK			= m_params->NTOPICS;
	m_nLabelNum		= m_params->NLABELS,
	m_dC			= m_params->INITIAL_C;
	m_dDeltaEll     = m_params->DELTA_ELL;
	m_dNuSqInv      = m_params->INITIAL_NUSQINV; //1 / m_dC;
	m_dAlpha 		= m_params->INITIAL_ALPHA / m_params->NTOPICS;
	m_dBeta 		= m_params->INITIAL_BETA; 
	m_nBurnIn       = m_params->BURN_IN;
	m_nMaxSamples   = m_params->MAX_SAMPLES;

	if (m_ptststats) delete m_ptststats;
	m_ptststats = new Stats();
	m_ptststats->new_stats(corpus, m_params);
	if (m_ptststats->nV > m_nV) 
		m_nV = m_ptststats->nV;
	m_dVbeta		= m_nV * m_dBeta;

	if(!m_dPr) m_dPr = new double[m_nK];

	const int infD	= m_ptststats->nD;
	int **infnw		= m_ptststats->nw;
	int *infnwsum	= m_ptststats->nwsum;
	int *zptr 		= NULL;
	int *infnd 		= NULL;
	int topic, term;
	for (int d = 0; d < infD; ++d) {
		Document& dc 	= corpus->docs[d];
		zptr				= m_ptststats->z[d];
		infnd				= m_ptststats->nd[d];
		for( int n = 0; n < dc.length; ++n) {
			term = dc.words[n];
			topic = (int)(myrand() * m_nK);
			*zptr++ = topic;
			infnw[topic][term] 	+= 1;
			infnwsum[topic]		+= 1;
			infnd[topic]			+= 1;
		}
	}
}

void MultiTaskGibbsMedLDA::init_inf(const char* model_file)
{
	load_model(model_file);

	Corpus* corpus = new Corpus();
#ifdef _REMOVE_OOV_
	corpus->read_data_oov(m_params->test_filename, m_nV);
#else
	corpus->read_data_binary(m_params->test_filename);
#endif
	init_inf(corpus);
}

void MultiTaskGibbsMedLDA::train()
{
	double dStartTime = get_runtime();
	Corpus *pTrC = m_ptrnstats->corpus;
	m_pIGsamplerG = new InverseGaussian(1, 1);

	// burn in stage
	printf(">> burn-in stage (iter %d): ", m_nBurnIn);
	for ( int it=0; it<m_nBurnIn; it++ ) {
		printf(".");
		double dtime1 = get_runtime();
		drawLambda();
		double dtime2 = get_runtime();
		m_dTrainTime_Lambda += (dtime2 - dtime1) / 100;

		drawEta();
		dtime1 = get_runtime();
		m_dTrainTime_Eta += (dtime1 - dtime2) / 100;

		drawZ_single_iter();
	}

	// sample stage (keep the samples)
	printf("\n>> burn-in done \n");
	printf(">> draw samples (iter %d): ", m_nMaxSamples);
	int iter = 0;
	for ( ; iter<m_nMaxSamples; iter++ ) {
		printf(".");
		double dtime1 = get_runtime();
		drawLambda();
		double dtime2 = get_runtime();
		m_dTrainTime_Lambda += (dtime2 - dtime1) / 100;

		drawEta();
		dtime1 = get_runtime();
		m_dTrainTime_Eta += (dtime1 - dtime2) / 100;

		for ( int y=0; y<m_nLabelNum; y++ ) {
			for ( int k=0; k<m_nK; k++ ) 
				m_dSampleEta[iter][y][k] = m_dEta[y][k];
		}
		drawZ_single_iter();
	}
	printf(">> draw samples done \n");

	// the average is the final weight
	for ( int y=0; y<m_nLabelNum; y++ ) {
		memset(m_dEta[y], 0, sizeof(double)*m_nK);
		for ( int n=0; n<m_nMaxSamples; n++ ) {
			addvec(m_dEta[y], m_dSampleEta[n][y], m_nK, 1.0/m_nMaxSamples);
		}
	}
	double dEndTime = get_runtime();
	m_dTrainTime = (dEndTime - dStartTime) / 100;

	char buf[1024];
	sprintf(buf, "%s/model.txt", m_params->dir);
	save_model(buf);

	sprintf(buf, "%s/lambda.txt", m_params->dir);
	save_lambda(buf);

	sprintf(buf, "%s/time.txt", m_params->dir);
	FILE *ftime = fopen(buf, "w");
	fprintf(ftime, "K: %d\n", m_nK);
	fprintf(ftime, "Total training time: %lf\n", m_dTrainTime);
	fclose(ftime);

	//printf("Final E-step\n");
	//double lhood = drawZ(iter);
	//printf("Final E-step Likelihood: %5.10f\n", lhood);
	//printf("Final E-step finished!\n");

	sprintf(buf, "%s/topwords.txt", m_params->dir);
	m_ptrnstats->save_topwords(buf, m_params);

	m_dTrainAcc = 0;
	for (int d = 0; d < m_ptrnstats->nD; ++d) {
		Document& dc = pTrC->docs[d];
		dc.predlabel = predict(m_ptrnstats, d);	
		if (dc.gndlabel_bk == dc.predlabel) ++m_dTrainAcc;
	}
	m_dTrainAcc /= m_ptrnstats->nD;
	printf("\n\n Double count accuracy: %.5f\n\n", m_dTrainAcc);
	sprintf(buf, "%s/double_count_accuracy.txt", m_params->dir);
	FILE* dfile = fopen(buf, "w");
	fprintf(dfile, "Double count accuracy: %.5f\n", m_dTrainAcc);
	fclose(dfile);

	// free memory
	delete m_pIGsamplerG;
}

void MultiTaskGibbsMedLDA::drawLambda()
{
	const int m_nD = m_ptrnstats->nD;
	double dMu;
	double cfactor = m_dC * m_dC;
	Corpus *pTrC = m_ptrnstats->corpus;

	for ( int y=0; y<m_nLabelNum; y++ ) { // for each category (i.e., each task)
		double *pLambdaInv = m_dLambdaInv[y];

		for ( int d=0; d<m_nD; d++ ) {    // for each document
			Document& dc = pTrC->docs[d];
			dc.gndlabel = adjust_gndlabel(dc.gndlabel_bk, y); // get the binary label of the document in task y

			dMu = m_dC / fabs( m_dDeltaEll - dc.gndlabel * discFunc( m_ptrnstats->nd[d], dc.length, m_dEta[y] ) );

			m_pIGsamplerG->reset( dMu, cfactor );
			pLambdaInv[d] = m_pIGsamplerG->sample();
		}
	}
}

double MultiTaskGibbsMedLDA::discFunc(int *zptr, const int &dlen, double *dEta) 
{
	double dval = 0;

	int k=m_nK;
	int *ndptr = zptr;
	double *etaptr = dEta;
	while(k--) {
		dval += (*etaptr++) * (*ndptr++);
	}
	//for ( int k=0; k<m_nK; k++ ) {
	//	dval += dEta[k] * zptr[k];
	//}
	dval /= dlen;

	return dval;
}

void MultiTaskGibbsMedLDA::drawEta()
{
	Corpus *pTrC = m_ptrnstats->corpus;
	Document *pDoc = NULL;

	for ( int y=0; y<m_nLabelNum; y++ ) {
		for ( int d=0; d<pTrC->num_docs; d++ ) { // get the binary label of each document in task y
			pDoc = &(pTrC->docs[d]);
			pDoc->gndlabel = adjust_gndlabel( pDoc->gndlabel_bk, y );
		}

		drawEta( m_dEta[y], m_dLambdaInv[y] ); // sample the classifier weight for task y
	}
}

void MultiTaskGibbsMedLDA::drawEta( double *dEta, const double *dLambdaInv )
{
	double *dCovInvPtr = NULL;
	double dFactor, dLocalFactor;

	// initialization
	memset(dEta, 0, sizeof(double)*m_nK);
	for ( int k=0; k<m_nK; k++ ) {
		memset(m_dCovInv[k], 0, sizeof(double)*m_nK);
		m_dCovInv[k][k] = m_dNuSqInv;
	}

	Corpus *pTrC = m_ptrnstats->corpus;
	Document *pDoc = NULL;
	for ( int d=0; d<pTrC->num_docs; d++ ) {
		pDoc = &(pTrC->docs[d]);
		int *cptr = m_ptrnstats->nd[d];

		// covariance
		dFactor = dLambdaInv[d] / (pDoc->length * pDoc->length);
		for ( int i=0; i<m_nK; i++ ) {
			if ( cptr[i] > 0 ) {
				dCovInvPtr = m_dCovInv[i];
				dLocalFactor = dFactor * cptr[i];
				for ( int j=0; j<m_nK; j++ ) {
					dCovInvPtr[j] += dLocalFactor * cptr[j];
				}
			}
		}

		// mean
		dFactor = pDoc->gndlabel * (m_dDeltaEll*dLambdaInv[d] + m_dC) / pDoc->length;
		for ( int i=0; i<m_nK; i++ ) {
			if ( cptr[i] > 0 ) 
				dEta[i] += cptr[i] * dFactor;
		}
	}

	///* Strategy #1: use the posterior mean as the 'sample' */
	//// matrix inverse
	//inverse( m_dCovInv, m_dCov, m_nK );	

	//for ( int k=0; k<m_nK; k++ ) {
	//	m_dEtabk_[k] = dotprod( m_dCov[k], m_dEta, m_nK );
	//}
	//for ( int k=0; k<m_nK; k++ ) {
	//	m_dEta[k] = m_dEtabk_[k];
	//}

	/* Strategy #2: draw random samples from multivariate Gaussian */
	inverse_cholydec(m_dCovInv, m_dCov, m_dLowerTriangle, m_nK);
	for ( int k=0; k<m_nK; k++ ) {
		m_dEtabk_[k] = dotprod( m_dCov[k], dEta, m_nK );
	}
	if ( m_pMVGaussian == NULL ) m_pMVGaussian = new MVGaussian();
	m_pMVGaussian->nextMVGaussianWithCholesky(m_dEtabk_, m_dLowerTriangle, dEta, m_nK);
}

double MultiTaskGibbsMedLDA::drawZ_single_iter(bool bLhood)
{
	const int m_nD = m_ptrnstats->nD;
	int *zptr = NULL;
	int *ndptr = NULL;
	bool log_space_flag = false;
	double gamma;
	int nlen;
	if ( !m_dDelta_ ) m_dDelta_ = new double[m_nK];

	// sample topic assignment Z
	for(int d = 0; d < m_nD; ++d)
	{
		Document& dc = m_ptrnstats->corpus->docs[d];
		nlen         = dc.length;
		zptr         = m_ptrnstats->z[d]; 
		ndptr        = m_ptrnstats->nd[d];

		double gamma = 1.0 / (double)(nlen);
		log_space_flag = compute_mrgterm_doc(d, dc.gndlabel_bk, nlen, m_mrg, gamma);
		gamma = gamma * (gamma - 1);

		if (log_space_flag) {
			for(int n = 0; n < nlen; ++n) {
				compute_delta(m_dDelta_, ndptr, zptr[n], nlen, d, gamma);
				zptr[n] = gibbs_sampling_logspace(m_ptrnstats, dc.words[n], m_mrg, m_dDelta_, d, n);
			}
		} else {
			for(int n = 0; n < nlen; ++n) {
				compute_delta(m_dDelta_, ndptr, zptr[n], nlen, d, gamma);
				zptr[n]	= gibbs_sampling(m_ptrnstats, dc.words[n], m_mrg, m_dDelta_, d, n);
			}
		}
	}

	double lhood = 0;
	if ( bLhood ) lhood = compute_lhood(m_ptrnstats);

	return lhood;
}

// an alternative implementation with pseudo-likelihood
double MultiTaskGibbsMedLDA::drawZ_single_iter2(bool bLhood)
{
	const int m_nD = m_ptrnstats->nD;
	int *zptr = NULL;
	int *ndptr = NULL;
	bool log_space_flag = false;
	double gamma, dfactor;
	int nlen, yorg;
	if ( !m_dDelta_ ) m_dDelta_ = new double[m_nK];
	if ( !m_dDiscValue_ ) m_dDiscValue_ = new double[m_nLabelNum];
	memset(m_mrg, 0, sizeof(double)*m_nK);

	// sample topic assignment Z
	for(int d = 0; d < m_nD; ++d)
	{
		Document& dc = m_ptrnstats->corpus->docs[d];
		nlen         = dc.length;
		zptr         = m_ptrnstats->z[d]; 
		ndptr        = m_ptrnstats->nd[d];
		yorg         = dc.gndlabel_bk;

		for ( int y=0; y<m_nLabelNum; y++ ) // compute the discriminantive function value for the document.
			m_dDiscValue_[y] = discFunc(ndptr, nlen, m_dEta[y]);

		for(int n = 0; n < nlen; ++n) {
			int topic = zptr[n];
			for ( int y=0; y<m_nLabelNum; y++ ) // discriminat function without the current word (and topic)
				m_dDiscValue_[y] -= m_dEta[y][topic] / nlen;

			// update the topic assignment
			compute_delta2(m_dDelta_, ndptr, zptr[n], nlen, d, yorg);
			zptr[n] = gibbs_sampling_logspace(m_ptrnstats, dc.words[n], m_mrg, m_dDelta_, d, n);

			// update the discriminat function value
			topic = zptr[n];
			for ( int y=0; y<m_nLabelNum; y++ ) // discriminat function with the current word (and topic)
				m_dDiscValue_[y] += m_dEta[y][topic] / nlen;
		}
	}

	double lhood = 0;
	if ( bLhood ) lhood = compute_lhood(m_ptrnstats);

	return lhood;
}

double MultiTaskGibbsMedLDA::drawZ(int em_it)
{
	char model_file[500];
	double lhood = 0;
	double lhood_old = 0;
	double converged = 1;

	printf("gibbs_conv: %.10lf\tmax_iters:%d\n", m_params->GIBBS_CONVERGED, m_params->NITERS);
	if (m_params->GIBBS_CONVERGED >= 0)
	{
		int iter = 0;
		while((iter < MIN_GIBBS_ITERATION) || (converged > m_params->GIBBS_CONVERGED) && (iter < m_params->NITERS) || (m_params->NITERS < 0))
		{
			printf("Iteration %d ..., ", iter);

			lhood = drawZ_single_iter(true);

			if ((m_params->ITER_SAVE_STEP > 0) && (iter % m_params->ITER_SAVE_STEP == 0)) {
				sprintf(model_file, "%s/model-%d-%d", m_params->dir, em_it, iter);
				save_model(model_file);
			}
			//lhood = compute_lhood(m_ptrnstats);
			converged = (lhood_old - lhood);
			converged = fabs(converged/lhood_old);
			printf("\tlhood: %f\tdelta: %f\n", lhood, converged);
			lhood_old = lhood;
			++iter;
		}
		printf("Done! \tlhood: %f\tdelta: %f\n", lhood, converged);
	}
	else;

	return lhood;
}

double MultiTaskGibbsMedLDA::inference(double *tAvgEnt)
{
	char buf[512];
	sprintf(buf, "%s/inf_single_likelihood.txt", m_params->dir);

	int nAcc	= 0;

	double converged  = 1;
	double lhood = 0, lhood_old = 0;
	double lhood_all = 0;
	double avgEnt = 0, dEnt, val;
	int *zptr 			= NULL;
	int zLen, n, iter, topic, k;

	const int infD 	= m_ptststats->nD;
	const int infV		= m_ptststats->nV;
	Document* docs 	= m_ptststats->corpus->docs;
	int** infnw			= m_ptststats->nw;
	int* infnwsum		= m_ptststats->nwsum;

	if (m_dPhi) {
		for(k = 0; k < m_nK; ++k) {
			if (m_dPhi[k]) delete[] m_dPhi[k];
			m_dPhi[k] = NULL;
		}
		delete[] m_dPhi;
	}
	m_dPhi = NULL;

	if (m_params->GIBBS_CONVERGED >= 0) 
		compute_phi(m_ptrnstats);
	int *suff_topics = new int[m_nK];

	time_t rawtime;
	struct tm *timeinfo;
	time(&rawtime);
	timeinfo = localtime(&rawtime);
	printf("K: %d >> Test start time: %s\n", m_nK, asctime(timeinfo));
	fflush(stdout);

	double dStartTime = get_runtime();
	printf("Prediction Inference ...\n");
	for(int d = 0; d < infD; ++d) {
		memset(infnwsum, 0, sizeof(int) * m_nK);
		for(k = 0; k < m_nK; ++k)
			memset(infnw[k], 0, sizeof(int)* infV);

		Document& dc = docs[d];
		zLen = dc.length; 
		zptr = m_ptststats->z[d];
		for(n = 0; n < zLen; ++n) {
			topic = zptr[n];
			infnw[topic][dc.words[n]] += 1;
			infnwsum[topic] += 1;
		}


		if (d % 1000 == 0) printf("Document %d\n", d);

		// burn-in stage
		iter = 0;
		lhood = 0;
		if (m_params->GIBBS_CONVERGED >= 0) {
			converged  = 1;
			lhood_old = 0;
		  	while((converged > m_params->GIBBS_CONVERGED) && (iter < m_params->NITERS) || (m_params->NITERS < 0)) {
		  		for(n = 0; n < zLen; ++n)
		  			zptr[n] = gibbs_sampling(m_ptrnstats, m_ptststats, d, n);
				compute_theta(m_ptststats, d);
		  		lhood = compute_lhood(&dc, m_dPhi, m_dTheta);
		  		converged = (lhood_old - lhood);
		  		converged = fabs(converged/lhood_old);
		  		lhood_old = lhood;
		  		++iter;
		  	}
		} else {
			for (iter = 0; iter < m_params->NITERS; ++iter)
			  	for(n = 0; n < zLen; ++n)
			  		zptr[n] = gibbs_sampling(m_ptrnstats, m_ptststats, d, n);
		}
		lhood_all += lhood;
		dc.lhood  = lhood;

		// predicting labels based on the sufficient statistics
		memset(suff_topics, 0, sizeof(int) * m_nK);
		for (iter = 0; iter < m_params->SAMPLING_LAG; ++iter) {
		  	for(n = 0; n < zLen; ++n) {
				topic = gibbs_sampling(m_ptrnstats, m_ptststats, d, n);
				zptr[n] = topic;
				suff_topics[topic] += 1;
			}
		}

		if (m_params->SAMPLING_LAG > 1){
			dEnt = 0;
			for(k = 0; k < m_nK; ++k) {
				val = (double)(suff_topics[k]/(double)m_params->SAMPLING_LAG + m_dAlpha)/(dc.length + m_nK * m_dAlpha);
				if (val > 0) dEnt -= val * log(val);
				m_dTheta[k] = val;
			}
			avgEnt += dEnt;
		} else if (m_params->SAMPLING_LAG == 1){
			double DV = (double)dc.length + m_nK * m_dAlpha;
			dEnt = 0;
			for(k = 0; k < m_nK; ++k) {
				val = ((double)suff_topics[k]+ m_dAlpha)/DV;
				dEnt -= val * log(val);
				m_dTheta[k] = val;
			}
			avgEnt += dEnt;
		} else ;

		predict(m_dTheta, dc);
		if (dc.predlabel == dc.gndlabel_bk ) ++nAcc;
	}
	m_dTestTime = (get_runtime() - dStartTime) / 100;

	sprintf(buf, "%s/time.txt", m_params->dir);
	FILE *ftime = fopen(buf, "a");
	fprintf(ftime, "Test start time: %s\n", asctime(timeinfo));
	time(&rawtime);
	timeinfo = localtime(&rawtime);
	fprintf(ftime, "Test end time: %s; Total: %.5f seconds \n", asctime(timeinfo), m_dTestTime);
	fclose(ftime);
	printf("K: %d >> Test end time: %s\n", m_nK, asctime(timeinfo));
	fflush(stdout);

	printf("Likilihood_ALL : %5.10f\n", lhood_all);
	printf("Prediction Inference FINISHED !\n");
	*tAvgEnt = avgEnt / infD;
	delete[] suff_topics;
	return (double) nAcc / infD;
}

int MultiTaskGibbsMedLDA::gibbs_sampling_logspace(Stats* ss, const int &term, const double* mrg, const double *delta, const int d, const int n)
{
	//Document& dc		= ss->corpus->docs[d];
	//const int term		= dc.words[n];
	int **m_nw			= ss->nw;
	int *m_nwsum		= ss->nwsum;
	int *ndptr			= ss->nd[d];
	int topic 			= ss->z[d][n];
	//double m_dVbeta		= m_nV * m_dBeta;

	m_nw[topic][term]   -= 1;
	m_nwsum[topic]		-= 1;
	ndptr[topic]		-= 1;

	for(int k = 0; k < m_nK; ++k)
		m_dPr[k] = log((m_nw[k][term] + m_dBeta) / (m_nwsum[k] + m_dVbeta) * (ndptr[k] + m_dAlpha)) + (mrg[k] + delta[k]);
	double logsum = log_sum(m_dPr, m_nK);

	m_dPr[0] = exp(m_dPr[0] - logsum);
	for(int k = 1; k < m_nK; ++k)
		m_dPr[k] = m_dPr[k-1] + exp(m_dPr[k] - logsum); 

	double rd = myrand() * m_dPr[m_nK-1];
	for(topic = 0; m_dPr[topic] < rd; ++topic);

	m_nw[topic][term]	+= 1;
	m_nwsum[topic]		+= 1;
	ndptr[topic]		+= 1;

	return topic;
}

int MultiTaskGibbsMedLDA::gibbs_sampling(Stats* ss, const int &term, const double* mrg, const double *delta, const int d, const int n)
{
	//Document& dc		= ss->corpus->docs[d];
	//const int term		= dc.words[n];
	int **m_nw			= ss->nw;
	int *m_nwsum		= ss->nwsum;
	int *ndptr			= ss->nd[d];
	int topic 			= ss->z[d][n];
	//double m_dVbeta		= m_nV * m_dBeta;

	m_nw[topic][term]   -= 1;
	m_nwsum[topic]		-= 1;
	ndptr[topic]		-= 1;

	m_dPr[0] = (m_nw[0][term] + m_dBeta) / (m_nwsum[0] + m_dVbeta) * (ndptr[0] + m_dAlpha) * exp(mrg[0] + delta[0]);
	for(int k = 1; k < m_nK; ++k)
		m_dPr[k] = m_dPr[k-1] + (m_nw[k][term] + m_dBeta) / (m_nwsum[k] + m_dVbeta) * (ndptr[k] + m_dAlpha) * exp(mrg[k] + delta[k]); 

	double rd = myrand() * m_dPr[m_nK-1];
	for(topic = 0; m_dPr[topic] < rd; ++topic);

	m_nw[topic][term]	+= 1;
	m_nwsum[topic]		+= 1;
	ndptr[topic]		+= 1;

	return topic;
}

bool MultiTaskGibbsMedLDA::compute_mrgterm_doc(const int d, const int yOrg, const int dlen, double* mrg, const double &gamma)
{
	double dval, eta, gndEllC = 0;

	bool flag = true;
	memset(mrg, 0, sizeof(double)*m_nK);
	double halfGamma = 0.5 * gamma;
	for ( int y=0; y<m_nLabelNum; y++ ) { // the marginal terms contributed from each task.
		double *dEta = m_dEta[y];
		double lambdaInv = m_dLambdaInv[y][d];
		double gammaLambda = gamma * lambdaInv;
		//double lambdaC = m_dC / lambdaInv;
		//int gndLabel = adjust_gndlabel(yOrg, y);
		if ( yOrg == y ) gndEllC = (m_dDeltaEll + m_dC/lambdaInv);
		else gndEllC = 0 - (m_dDeltaEll + m_dC/lambdaInv);

		for(int k = 0; k < m_nK; ++k) {
			eta = dEta[k];
			mrg[k] += gammaLambda * eta * ( gndEllC - halfGamma * eta );
		}
	}
	return flag;
}

// discriminant function value with word n removed.
void MultiTaskGibbsMedLDA::compute_delta(double *dDelta, int *ndptr, int topic, const int &dlen, const int &d, const double &gamma)
{
	memset(dDelta, 0, sizeof(double)*m_nK);

	if ( dlen > 1 ) {
		ndptr[topic] --;
		for ( int y=0; y<m_nLabelNum; y++ ) {
			double *dEta = m_dEta[y];
			double dfactor = gamma * m_dLambdaInv[y][d];

			// the delta value for each task (category)
			double dval = discFunc(ndptr, dlen-1, dEta);

			// add to the overall delta
			dval *= dfactor;
			for ( int k=0; k<m_nK; ++k) {
				dDelta[k] += dEta[k] * dval;
			}
		}
		ndptr[topic] ++;
	}
}

void MultiTaskGibbsMedLDA::compute_delta2(double *dDelta, int *ndptr, int topic, const int &dlen, const int &d, const int &yOrg)
{
	double discVal = 0, zetaVal = 0, lambdaInv = 0, deltaVal = 0;
	for ( int k=0; k<m_nK; k++ ) {
		deltaVal = 0;
		for ( int y=0; y<m_nLabelNum; y++ ) {
			discVal = m_dDiscValue_[y] + m_dEta[y][k] / dlen; // discriminat function with current word (and possible topic assignment k)

			if ( yOrg == y ) zetaVal = m_dDeltaEll - discVal;
			else zetaVal = m_dDeltaEll + discVal;

			lambdaInv = m_dLambdaInv[y][d];
			deltaVal += (1 + zetaVal * lambdaInv) * (1/lambdaInv + zetaVal);
		}
		dDelta[k] = -0.5 * deltaVal;
	}
}

void MultiTaskGibbsMedLDA::compute_theta(Stats* ss, int d)
{
	if (!m_dTheta)
		m_dTheta = new double[m_nK];	

	const double DIV = m_dAlpha * m_nK + ss->corpus->docs[d].length;
	int* nd_ptr = ss->nd[d];
	double* theta_ptr = m_dTheta;
	int k = m_nK;
	while(k--)
		*theta_ptr++ = double (*nd_ptr++ + m_dAlpha) / DIV; 
}

bool MultiTaskGibbsMedLDA::compute_phi(Stats* st)
{
	if ( st->nV > m_nV) {
		printf(" ERROR> V illeagle\n");	
		return false;
	}

	if(!m_dPhi) {
		m_dPhi = new double*[m_nK];
		for(int k = 0; k < m_nK; ++k)
			m_dPhi[k] = new double[m_nV];
	}

	//const double m_dVbeta	= m_nV * m_dBeta;
	int t, nwsum;
	double* phi_ptr = NULL;
	int* nw_ptr = NULL;
	for(int k = 0; k < m_nK; ++k)
	{
		phi_ptr 	= m_dPhi[k];
		nw_ptr	= st->nw[k];
		const double div	= st->nwsum[k] + m_dVbeta;
		for (t = 0; t < st->nV; ++t)
			*phi_ptr++ = double (*nw_ptr++ + m_dBeta) / div; 
		for (; t < m_nV; ++t)
			*phi_ptr++ = m_dBeta / div;
	}
	return true;
}

double MultiTaskGibbsMedLDA::compute_lhood(Stats* st)
{
	double lhood = 0.0;
	double dlhood = 0, dlword, theta_k;
	int term, k, n;
	int *nd_ptr;
	bool log_space_flag = false;

	compute_phi(st);
	for(int d = 0; d < st->nD; ++d)
	{
		dlhood = 0;
		Document& dc = st->corpus->docs[d];	
		log_space_flag = compute_mrgterm_doc(d, dc.gndlabel, dc.length, m_mrg, 1.0/dc.length);
		compute_theta(st, d);
		for(k = 0; k < m_nK; ++k)
			dlhood += m_dTheta[k] * m_mrg[k] * dc.length;

		if (log_space_flag) {
			for(k = 0; k < m_nK; ++k)
				m_dTheta[k] = log(m_dTheta[k]) + m_mrg[k];
			for (n = 0; n < dc.length; ++n) {
				term = dc.words[n];	
				dlword = 0;
				for(k = 0; k < m_nK; ++k)
					dlword = log_sum(dlword, log(m_dPhi[k][term]) + m_dTheta[k]);
				dlhood -= dlword;
			}
		} else {
			for(k = 0; k < m_nK; ++k)
				m_dTheta[k] *= exp(m_mrg[k]);
			for (n = 0; n < dc.length; ++n) {
				dlword = 0;
				term = dc.words[n];	
				for(k = 0; k < m_nK; ++k)
					dlword += m_dPhi[k][term] * m_dTheta[k];
				dlhood -= log(dlword);
			}
		}
		lhood += dlhood;
	}
	return lhood;
}


double MultiTaskGibbsMedLDA::compute_lhood(Document* dc, double** phi, double* theta) {
	double dlhood = 0, dlword;
	int term;
	for (int n = 0; n < dc->length; ++n)
	{
		dlword = 0;
		term = dc->words[n];	
		for(int k = 0; k < m_nK; ++k)
			dlword += phi[k][term] * theta[k];
		dlhood -= log(dlword);
	}
	return dlhood;
}

int MultiTaskGibbsMedLDA::gibbs_sampling(Stats* trn_ss, Stats* tst_ss, const int d, const int n)
{
	//double m_dVbeta		= m_dBeta * m_nV;

	Document& dc		= tst_ss->corpus->docs[d];
	const int term		= dc.words[n];
	int topic 			= tst_ss->z[d][n];
	int **m_infnw		= tst_ss->nw;
	int *m_infnwsum	= tst_ss->nwsum;
	int **m_infnd		= tst_ss->nd;

	int **m_nw			= trn_ss->nw;
	int *m_nwsum		= trn_ss->nwsum;
	int **m_nd			= trn_ss->nd;


	m_infnw[topic][term] -= 1;
	m_infnwsum[topic]		-= 1;
	m_infnd[d][topic]		-= 1;

	if ( term < trn_ss->nV)
	{
		m_dPr[0] = (m_nw[0][term] + m_infnw[0][term] + m_dBeta) / (m_nwsum[0] + m_infnwsum[0] + m_dVbeta) * (m_infnd[d][0] + m_dAlpha);
		for(int k = 1; k < m_nK; ++k)
			m_dPr[k] = m_dPr[k-1] + (m_nw[k][term] + m_infnw[k][term] + m_dBeta) / (m_nwsum[k] + m_infnwsum[k] + m_dVbeta) * (m_infnd[d][k] + m_dAlpha);
	}
	else 
	{
		m_dPr[0] = (m_infnw[0][term] + m_dBeta) / (m_infnwsum[0] + m_dVbeta) * (m_infnd[d][0] + m_dAlpha); 
		for(int k = 1; k < m_nK; ++k)
			m_dPr[k] = m_dPr[k-1] + (m_infnw[k][term] + m_dBeta) / (m_infnwsum[k] + m_dVbeta) * (m_infnd[d][k] + m_dAlpha);
	}

	double rd = myrand() * m_dPr[m_nK-1];
	for(topic = 0; m_dPr[topic] < rd; ++topic);

	m_infnw[topic][term]	+= 1;
	m_infnwsum[topic]		+= 1;
	m_infnd[d][topic]		+= 1;

	return topic;
}

void MultiTaskGibbsMedLDA::predict(double *theta, Document &dc)
{
	int pred_label = -1;		
	double dMaxScore = 0;
	for ( int y=0; y<m_nLabelNum; y++ ) {
		double dScore = 0;										
		int k = m_nK;
		double *ptr = theta;
		double *eta_ptr = m_dEta[y];
		while(k--)
			dScore += (*eta_ptr++) * (*ptr++); 
		dScore -= m_dB;

		if ( pred_label == -1 || dMaxScore < dScore ) {
			dMaxScore = dScore;
			pred_label = y;
		}
	}

	dc.predlabel = pred_label;
	dc.predscore = dMaxScore;
}

//int MultiTaskGibbsMedLDA::predict(int* topics, int doc_len, int num_iters)
//{
//	const double Kalpha = m_nK * m_dAlpha;
//	double* eta_ptr = m_dEta;
//
//	double dScore = 0;
//	int k = m_nK;
//	int *nd_ptr = topics;
//	while(k--)
//		dScore += (*eta_ptr++) * (double (*nd_ptr++) / double(num_iters) + m_dAlpha) / (doc_len + Kalpha); 
//	dScore -= m_dB;
//
//	int pred_label = -1;		
//	if ( dScore > 0 ) pred_label = 1;
//
//	return pred_label;
//}

int MultiTaskGibbsMedLDA::predict(Stats* ss, const int d)
{
	const int len	= ss->corpus->docs[d].length;
	const double Kalpha = m_nK * m_dAlpha;

	int pred_label = -1;		
	double dMaxScore = 0;
	for ( int y=0; y<m_nLabelNum; y++ ) {
		double dScore = 0;										
		int k = m_nK;
		int *nd_ptr = ss->nd[d];
		double* eta_ptr = m_dEta[y];
		while(k--)
			dScore += (*eta_ptr++) * (*nd_ptr++ + m_dAlpha) / (len + Kalpha);
		dScore -= m_dB;

		if ( pred_label == -1 || dMaxScore < dScore ) {
			dMaxScore = dScore;
			pred_label = y;
		}
	}
	
	return pred_label;
}

void MultiTaskGibbsMedLDA::outputLowDimData(Stats* ss, char *filename)
{
	double val;
	double Kalpha = m_nK * m_dAlpha; 
	FILE *fileptr = fopen(filename, "w");
	for ( int d = 0; d < m_ptrnstats->nD; d++ ) 
	{
		Document& dc = ss->corpus->docs[d];
		fprintf(fileptr, "%d %d", m_nK, dc.gndlabel);

		int* zz = ss->nd[d];
		for ( int k = 0; k < m_nK; k++ ) {
			val = (double(zz[k]) + m_dAlpha) / (double(dc.length) + Kalpha);
			fprintf(fileptr, " %d:%.10f", k, val);
		}
		fprintf(fileptr, "\n");
	}
	fclose(fileptr);
}

void MultiTaskGibbsMedLDA::save_results(double tAvgEnt)
{
	printf(">>> Saving results & evaluating performace\n");
	char buf[512];

	int nAcc = 0;
	int total = 0;
	double lhood = 0;
	for(int d = 0; d < m_ptststats->nD; ++d)
	{
		Document &dc = m_ptststats->corpus->docs[d];
		dc.gndlabel = dc.gndlabel_bk; // restore the gnd labels for saving.
		lhood	+= dc.lhood;
		total += dc.length;
		if (dc.gndlabel == dc.predlabel)
			++nAcc;
	}
	double dAcc = (double) nAcc / m_ptststats->nD;
	double perwordlhood = lhood / double(total);
	printf("Accuracy: %5.5f\n", dAcc);

	sprintf(buf, "%s/evl-performance.txt", m_params->dir);
	FILE* fileptr = fopen(buf, "w");
	fprintf(fileptr, "num docs: %d\n", m_ptststats->nD);
	fprintf(fileptr, "num terms: %d\n", m_nV);
	fprintf(fileptr, "num topics: %d\n", m_nK);
	fprintf(fileptr, "cost parameter l: %d\n", (int)m_dDeltaEll);
	fprintf(fileptr, "Alpha: %5.10f\n", (m_dAlpha*m_nK)); 
	//fprintf(fileptr, "Sample Lag: %d\n", m_params->SAMPLING_LAG);
	fprintf(fileptr, "C:	%5.10f\n\n", m_dC);
	fprintf(fileptr, "Accuracy: %5.10f\n", dAcc);
	fprintf(fileptr, "Likelihood: %5.10f\n", lhood);
	fprintf(fileptr, "PerWordLikelihood: %5.10f\n", perwordlhood);
	fprintf(fileptr, "testAverageEntropy: %5.10f\n", tAvgEnt);
	fclose(fileptr);

	sprintf(buf, "%s/pred_wordassigned.txt", m_params->dir);
	m_ptststats->save_wordassigned(buf);
	sprintf(buf, "%s/pred_results.txt", m_params->dir);
	m_ptststats->save_label_only(buf);

	sprintf(buf, "%s/gml_%dtopic_test.txt", m_params->dir, m_params->NTOPICS);
	FILE* fout = fopen(buf, "w");
	if (!fout) return;
	fprintf(fout, "%d\n", m_ptststats->nD);
	for(int d = 0; d < m_ptststats->nD; ++d) {
		Document &dc = m_ptststats->corpus->docs[d];
		fprintf(fout, "%d %d", m_nK, dc.gndlabel);
		compute_theta(m_ptststats, d);	
		for (int k = 0; k < m_nK; ++k)
			fprintf(fout, " %d:%.10f", k, m_dTheta[k]);
		fprintf(fout, "\n");
	}
	fclose(fout);

	// overall-results
	FILE *fres = fopen(m_params->res_filename, "a");
	fprintf(fres, "(k: %d, c: %.3f, NuSqInv: %.3f, alpha: %.3f, beta: %.3f, ell: %d, em-iter: %d, burn-in: %d, samples: %d) acc: %.4f, trn-acc: %.4f, trn-time: %.4f, tst-time: %.4f, eta: %.4f, lambda: %.4f\n", 
		m_nK, m_dC, m_dNuSqInv, m_params->INITIAL_ALPHA, m_params->INITIAL_BETA,
		(int)m_dDeltaEll, m_params->EM_MAX_ITER, m_params->BURN_IN, m_params->MAX_SAMPLES,
		dAcc, m_dTrainAcc, m_dTrainTime, m_dTestTime, m_dTrainTime_Eta/m_dTrainTime, m_dTrainTime_Lambda/m_dTrainTime);
	fflush(fres);
	fclose(fres);

	printf(">>> Saving results & evaluating performace finished! \n");
}

void MultiTaskGibbsMedLDA::save_model_property(Stats* st, Params *param)
{
	char buf[512];	
	sprintf(buf, "%s/model_property.txt", param->dir);
	FILE *file = fopen(buf, "w");
	int d, k, t;
	int nZero = 0;
	compute_phi(st);
	for(k = 0; k < m_nK; ++k)
		for(t = 0; t < m_nV; ++t)
			if (m_dPhi[k][t] < NUMERICAL_ZERO) nZero++;
	double dZeroPercent = double(nZero) / (m_nK * m_nV);
	fprintf(file, "ZeroPercentage: %.10f\n", dZeroPercent);

	double *theta = NULL;
	double dEntropy = 0;
	double entropy = 0;
	double val;

	sprintf(buf, "%s/theta_train.txt", param->dir);
	FILE *thetaFile = fopen(buf, "w");

	for(d = 0; d < st->nD; ++d){
		compute_theta(st, d);
		theta = m_dTheta;
		k = m_nK;
		dEntropy = 0;
		while(k--){
			val = *theta++;
			fprintf(thetaFile, "%.10f ", val);
			if (val > 0)
				dEntropy -= val * log(val);
		}
		fprintf(thetaFile, "\n");
		entropy += dEntropy;
	}
	fclose(thetaFile);
	entropy /= double(st->nD);
	fprintf(file, "AverageEntropy: %.10f\n", entropy);
	fclose(file);
}

void MultiTaskGibbsMedLDA::save_model(const char* model_file)
{
	printf(">>> Saving model ...\n");
	FILE* file = fopen(model_file, "w");
	if (!file) {
		printf("ERROR> Cannot open model file error !");	
		return;
	}

	fprintf(file, "%f\n", m_dC);

	fprintf(file, "%.10f\n", m_dB);
	for ( int y=0; y<m_nLabelNum; y++ ) {
		double* ptr = m_dEta[y];
		int N = m_nK;
		while(N--)
			fprintf(file, "%5.10f ", *ptr++); 
		fprintf(file, "\n");
	}
	fprintf(file, "\n");

	m_ptrnstats->save_wordassigned(file);
	if ( !m_dPhi ) compute_phi( m_ptrnstats );
	for(int i = 0; i < m_nK; ++i){
		for(int j = 0; j < m_nV; ++j){
			fprintf(file, "%.10f ", m_dPhi[i][j]);
		}
		fprintf(file, "\n");
	}
	fclose(file);
	printf(">>> Savning model finished !\n");
}

void MultiTaskGibbsMedLDA::load_model(const char* model_file)
{
	printf(">>> Loading model ...\n");
	FILE* file = fopen(model_file, "r");
	if (!file) {
		printf("ERROR> Cannot open model file error !");	
		return;
	}

	// Loading C
	fscanf(file, "%lf\n", &(m_params->INITIAL_C));
	fprintf(file, "%lf\n", m_dB);

	// Loading eta (svm classifier)
	int N = m_params->NTOPICS;
	m_dEta = new double*[m_params->NTOPICS];
	for ( int y=0; y<m_params->NTOPICS; y++ ) {
		m_dEta[y] = new double[N];
		for(int i = 0; i < N; ++i)
			fscanf(file, "%lf", &m_dEta[y][i]);
	}

	// Loading MultiTaskGibbsMedLDA model
	if (m_ptrnstats) delete m_ptrnstats;
	m_ptrnstats = new Stats();
	m_ptrnstats->load(file, m_params);
	m_nV = m_ptrnstats->nV;

	fclose(file);
	printf(">>> Loading model finished! \n");
}

void MultiTaskGibbsMedLDA::save_lambda(const char* lambda_file)
{
	FILE *fptr = fopen(lambda_file, "w");
	for ( int d=0; d<m_ptrnstats->corpus->num_docs; d++ ) {
		for ( int y=0; y<m_nLabelNum; y++ ) 
			fprintf(fptr, "%.5f ", m_dLambdaInv[y][d]);
		fprintf(fptr, "\n");
	}
	fclose(fptr);
}
