#ifndef _GIBBS_MED_LDA_MT_H
#define _GIBBS_MED_LDA_MT_H

#include "defs.h"
#include "corpus.h"
#include "params.h"
#include <stdlib.h>
#include <stdio.h>
#include <iostream>
#include <string.h>
#include <time.h>

#include "InverseGaussian.h"
#include "MVGaussian.h"
#include "stats.h"
using namespace std;

// The multi-task GibbsMedLDA for multi-class classification

class MultiTaskGibbsMedLDA
{
public:
	MultiTaskGibbsMedLDA(Params *params): m_params(params) 
	{ 
		m_pMVGaussian = NULL;
		if (!m_params)
		{
			printf("ERROR> NO params error!\n");
			exit(1);
		}
		set_default_values();
	}
	~MultiTaskGibbsMedLDA();

/*
*	sharing	
*/

public:
	void set_default_values();

	void save_model(const char* model_file);
	void load_model(const char* model_file);
	void save_model_property(Stats* st, Params *param);
	void save_lambda(const char* lambda_file);

	int predict(Stats* ss, const int d);
	//int predict(int* topics, int doc_len, int num_iters);
	void predict(double* theta, Document &dc);

private:
	void compute_theta(Stats* ss, int d);
	bool compute_phi(Stats* st);

/*
*	training 
*/

public:
	void init_est(int refLabel = -1); 
	void init_est(Corpus* corpus);
	void init_estc(const char* prev_state_file);
	void run_em();

	void train();

private:
	void drawEta();
	void drawEta( double *dEta, const double *dLambdaInv );
	void drawLambda();
	double drawZ_single_iter(bool bLhood = false);
	double drawZ_single_iter2(bool bLhood = false);
	double drawZ(int em_it);

	int gibbs_sampling(Stats* ss, const int &term, const double* mrg, const double *delta, const int d, const int n);
	int gibbs_sampling_logspace(Stats* ss, const int &term, const double* mrg, const double *delta, const int d, const int n);
	bool compute_mrgterm_doc(const int d, const int y, const int dlen, double* mrg, const double &gamma);
	void compute_delta(double *dDelta, int *ndptr, int topic, const int &dlen, const int &d, const double &gamma);
	void compute_delta2(double *dDelta, int *ndptr, int topic, const int &dlen, const int &d, const int &yOrg);
	double compute_lhood(Stats* st);

	void outputLowDimData(Stats* ss, char *filename);

/*
*	 prediction
*/

public:
	void init_inf(Corpus* corpus);
	void init_inf(const char* model_file);
	double inference(double *tAvgEnt);
	void save_results(double tAvgEnt);
	//void inference_semi();

private:
	int gibbs_sampling(Stats* trn_ss, Stats* tst_ss, const int d, const int n);
	double compute_lhood(Document* doc, double** phi, double* theta);
	//bool compute_phi(Stats* st, Stats* sp);
	//double compute_lhood(Stats* st, Stats* sp);

	double discFunc(int* zptr, const int &dlen, double *dEta);
	inline int adjust_gndlabel(const int &orgLabel, const int &refLabel) {
    if ( orgLabel == refLabel ) return 1;
    else return -1;
  }

public:
	Params* m_params;
	Stats* m_ptrnstats;
	Stats* m_ptststats;

	int m_nK;
	int m_nV;
	int m_nLabelNum; // number of classes, |C|
	double m_dC;
	double m_dAlpha; 
	double m_dBeta;
	double m_dVbeta;
	double m_dNuSqInv;
	double m_dDeltaEll;

	double **m_dEta; // #labels * #topics
	double **m_dCov;
	double **m_dCovInv;
	double **m_dLowerTriangle;

	double *m_dPr; // temp variable for sampling
	double *m_mrg; // 1 X #topics

	double *m_dTheta; // 1 \times #topics
	double **m_dPhi; // only used when calculating likelihood;
						  // #topic * #terms

	double m_dB;    // bias of SVM solver

	/*
	* intermediate variables for training only
	*/
	double **m_dLambdaInv; // #labels * #documents
	InverseGaussian *m_pIGsamplerG;
	double *m_dEtabk_;
	double *m_dDelta_;
	double *m_dDiscValue_;
	double ***m_dSampleEta;
	int m_nMaxSamples;
	int m_nBurnIn;
	double m_dTrainAcc;
	double m_dTrainTime;
	double m_dTrainTime_Eta;
	double m_dTrainTime_Lambda;
	double m_dTestTime;

	// the multivariante Gaussian
	MVGaussian *m_pMVGaussian;
};

#endif
