#include "defs.h"
#include "utils.h"
#include "cokus.h"
#include "gibbsMedLDA.h"
#include "gibbsMedLDA-MT.h"
#include "params.h"

#include <sys/stat.h>
#include <sys/types.h>
#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>

using namespace std;

double inner_cv_C(Corpus* corpus, Params* param);
double inner_cv_alpha(Corpus* corpus, Params* param);
double save_results_multi(double **predScore, int *gndLabel, int nTstData, double dTrainTime, double dTestTime, Params *param);

int main(int argc, char** argv)
{
	seedMT( time(NULL) );

	if (argc > 1)
	{
		Params param;
		param.read_settings(argv[5]);
		param.NTOPICS = atoi(argv[2]);
		param.NLABELS = atoi(argv[3]);
		param.INITIAL_ALPHA = atof(argv[4]);
		param.INITIAL_C = atof(argv[6]);
		param.EM_MAX_ITER = atoi(argv[7]);
		param.NFOLDS  = atoi(argv[8]);
		param.DELTA_ELL = atof(argv[9]);
		int nRunTime = 1;
		if ( argc > 10 ) nRunTime = atoi(argv[10]);

		printf("K: %d, Alpha: %.3f, C: %.3f, SampleLag: %d\n", param.NTOPICS, param.INITIAL_ALPHA, param.INITIAL_C, param.SAMPLING_LAG);
		sprintf(param.dir, "GML-K%d-A%d-L%d-C%d-F%d", param.NTOPICS, (int)param.INITIAL_ALPHA,(int)param.SAMPLING_LAG,  (int)param.INITIAL_C, param.NFOLDS);
		//make_directory(param.dir);
		mkdir(param.dir, 0777);
		sprintf(param.model_filename, "%s/model-final.txt", param.dir);

		if ( strcmp(argv[1], "-estinf") == 0 ) 
		{
			double* acc = new double[nRunTime];
			double* trnAcc = new double[nRunTime];
			double* trnTime = new double[nRunTime];
			double* trnTimeEta = new double[nRunTime];
			double* trnTimeLambda = new double[nRunTime];
			double* tstTime = new double[nRunTime];
			FILE *fsptr = fopen("sensitivity_analysis.txt", "a");
			fprintf(fsptr, "=== sensitivity to ell ===\n");
			for ( int i=0; i<25; i++ ) {
				//param.NTOPICS = (i+1) * 5;
				//param.INITIAL_ALPHA = 0.00001 * pow(10, 0.25*i); 
				param.DELTA_ELL = pow((1 + i), 2.0);
				for ( int m=0; m<nRunTime; m++ ) {
					if (param.INNER_CV_C) {
						Corpus* corpus = new Corpus();
						corpus->read_data_binary(param.train_filename);
						corpus->shuffle();
						param.INITIAL_C = inner_cv_C(corpus, &param);
						printf("\n\nBest C: %5.10f\n\n", param.INITIAL_C);
						delete corpus;
					}
					if (param.INNER_CV_ALPHA) {
						Corpus* corpus = new Corpus();
						corpus->read_data_binary(param.train_filename);
						corpus->shuffle();
						param.INITIAL_ALPHA = inner_cv_alpha(corpus, &param);
						printf("\n\nBest ALPHA: %5.10f\n\n", param.INITIAL_ALPHA);
						delete corpus;
					}

					GibbsMedLDA model(&param);
					model.init_est();
					if ( param.EM_MAX_ITER > 0 ) 
						model.run_em();
					else 
						model.train();
					model.save_model(param.model_filename);

					Corpus *test_corpus = new Corpus();
					// Read data and remove unseen words
					test_corpus->read_data_oov(param.test_filename, model.m_nV);
					model.init_inf(test_corpus);
					double tAvgEnt = 0;
					double dAcc = model.inference(&tAvgEnt);
					model.save_results(tAvgEnt);
					printf("Accuracy: %5.10f\n", dAcc);
					printf("testAverageEntropy: %5.10f\n", tAvgEnt);

					acc[m]				=  dAcc;
					trnAcc[m]			=  model.m_dTrainAcc;
					trnTime[m]			=  model.m_dTrainTime;
					trnTimeEta[m]		=  model.m_dTrainTime_Eta;
					trnTimeLambda[m]	=  model.m_dTrainTime_Lambda;
					tstTime[m]			=  model.m_dTestTime;
				}
				if ( nRunTime > 1 ) {
					double dAvgAcc = 0, dAvgTrnAcc = 0, dAvgTrnTime = 0, dAvgTstTime = 0, dAvgTrnTimeEta = 0, dAvgTrnTimeLambda = 0;
					double dStdAcc				= compute_mean_var( dAvgAcc, acc, nRunTime );
					double dStdTrnAcc			= compute_mean_var( dAvgTrnAcc, trnAcc, nRunTime);
					double dStdTrnTime			= compute_mean_var( dAvgTrnTime, trnTime, nRunTime );
					double dStdTrnTimeEta		= compute_mean_var( dAvgTrnTimeEta, trnTimeEta, nRunTime );
					double dStdTrnTimeLambda	= compute_mean_var( dAvgTrnTimeLambda, trnTimeLambda, nRunTime );
					double dStdTstTime			= compute_mean_var( dAvgTstTime, tstTime, nRunTime );
					printf("avg-acc: %.4f; %d\n", dAvgAcc, nRunTime);

					FILE *fptr = fopen(param.res_filename, "a");
					fprintf(fptr, "(k: %d, c: %.3f, NuSqInv: %.3f, alpha: %.3f, beta: %.3f, ell: %d, em-iter: %d, burn-in: %d, samples: %d) avg-acc: %.4f, std: %.4f; avg-train-time: %.4f, std: %.4f; avg-test-time: %.4f; std: %.4f\n", 
						param.NTOPICS, param.INITIAL_C, param.INITIAL_NUSQINV, param.INITIAL_ALPHA, param.INITIAL_BETA,
						(int)param.DELTA_ELL, param.EM_MAX_ITER, param.BURN_IN, param.MAX_SAMPLES,
						dAvgAcc, sqrt(dStdAcc), dAvgTrnTime, sqrt(dStdTrnTime), dAvgTstTime, sqrt(dStdTstTime) );
					fflush(fptr);
					fclose(fptr);

					//fprintf(fsptr, "%d\t%.10f\t%.4f\t%.4f\t%.4f\t%.4f\t%.4f\t%.4f\n", param.NTOPICS, param.INITIAL_ALPHA, dAvgAcc, sqrt(dStdAcc), dAvgTrnTime, sqrt(dStdTrnTime), dAvgTstTime, sqrt(dStdTstTime));
					fprintf(fsptr, "%d\t%.10f\t%.4f\t%.4f\t%.4f\t%.4f\t%.4f\t%.4f\n", 
						param.NTOPICS, param.DELTA_ELL, 
						dAvgAcc, sqrt(dStdAcc), dAvgTrnTime, sqrt(dStdTrnTime), dAvgTstTime, sqrt(dStdTstTime));
					fflush(fsptr);
				}
			}
			fclose(fsptr);
			delete[] acc;
			delete[] trnTime;
			delete[] tstTime;
		}
		else if ( strcmp(argv[1], "-multic") == 0 ) // 1 vs all strategy for multi-class
		{
			double* acc = new double[nRunTime];
			double* trnTime = new double[nRunTime];
			double* tstTime = new double[nRunTime];
			double* maxTrnTime = new double[nRunTime];
			FILE *fsptr = fopen("sensitivity_20ng_one-vs-all.txt", "a");
			fprintf(fsptr, " === burn-in ===\n");
			for ( int i=1; i<6; i++ ) {
				param.NTOPICS = (i+1) * 10;
				//param.INITIAL_ALPHA = (i + 1) * 10;
				for ( int n=0; n<nRunTime; n++ ) {

					int nTstData = 0;
					double **predScore = new double*[param.NLABELS];
					int *gndLabel = NULL;
					double dTrainTime = 0, dTestTime = 0, dMaxTrnTime = 0;
					for ( int m=0; m<param.NLABELS; m++ ) {
						printf(">>> class label %d \n", m+1);
						sprintf(param.model_filename, "%s/model-final_c%d.txt", param.dir, m);
						GibbsMedLDA model(&param);
						model.init_est(m);
						if ( param.EM_MAX_ITER > 0 ) 
							model.run_em();
						else 
							model.train();
						model.save_model(param.model_filename);
						dTrainTime += model.m_dTrainTime;
						dMaxTrnTime = model.m_dTrainTime > dMaxTrnTime ? model.m_dTrainTime : dMaxTrnTime;

						Corpus *test_corpus = new Corpus();
						// Read data and remove unseen words
						test_corpus->read_data_oov(param.test_filename, model.m_nV, m);
						model.init_inf(test_corpus);
						double tAvgEnt = 0;
						double dAcc = model.inference(&tAvgEnt);
						printf("Accuracy: %5.10f\n", dAcc);
						printf("testAverageEntropy: %5.10f\n", tAvgEnt);
						dTestTime += model.m_dTestTime;

						// copy the predicted score
						nTstData = test_corpus->num_docs;
						predScore[m] = new double[nTstData];
						if ( !gndLabel ) gndLabel = new int[nTstData];
						for ( int d=0; d<nTstData; d++ ) {
							predScore[m][d] = test_corpus->docs[d].predscore;
							gndLabel[d] = test_corpus->docs[d].gndlabel_bk;
						}
					}
					double dMultiAcc = save_results_multi(predScore, gndLabel, nTstData, dTrainTime, dTestTime, &param);

					delete[] gndLabel;
					for ( int m=0; m<param.NLABELS; m++ ) delete[] predScore[m];
					delete[] predScore;

					// average over multiple runs
					acc[n]			= dMultiAcc;
					trnTime[n]		= dTrainTime;
					tstTime[n]		= dTestTime;
					maxTrnTime[n]	= dMaxTrnTime;
				}
				if ( nRunTime > 1 ) {
					double dAvgAcc = 0, dAvgTrnTime = 0, dAvgTstTime = 0, dAvgMaxTrnTime = 0;
					double dStdAcc			= compute_mean_var( dAvgAcc, acc, nRunTime );
					double dStdTrnTime		= compute_mean_var( dAvgTrnTime, trnTime, nRunTime);
					double dStdTstTime		= compute_mean_var( dAvgTstTime, tstTime, nRunTime);
					double dStdMaxTrnTime	= compute_mean_var( dAvgMaxTrnTime, maxTrnTime, nRunTime );
					printf("avg-acc: %.4f; %d\n", dAvgAcc, nRunTime);

					FILE *fptr = fopen(param.res_filename, "a");
					fprintf(fptr, "(k: %d, c: %.3f, NuSqInv: %.3f, alpha: %.3f, beta: %.3f, ell: %d, em-iter: %d, burn-in: %d, samples: %d) avg-acc: %.4f, std: %.4f; avg-train-time: %.4f, std: %.4f; avg-test-time: %.4f; std: %.4f; parallel-avg-tr-time: %.4f; std: %.4f\n", 
						param.NTOPICS, param.INITIAL_C, param.INITIAL_NUSQINV, param.INITIAL_ALPHA, param.INITIAL_BETA,
						(int)param.DELTA_ELL, param.EM_MAX_ITER, param.BURN_IN, param.MAX_SAMPLES,
						dAvgAcc, sqrt(dStdAcc), dAvgTrnTime, sqrt(dStdTrnTime), dAvgTstTime, sqrt(dStdTstTime), dAvgMaxTrnTime, sqrt(dStdMaxTrnTime) );
					fflush(fptr);
					fclose(fptr);

					fprintf(fsptr, "%d:%d\t%.4f\t%.4f\t%.4f\t%.4f\n", param.NTOPICS, param.BURN_IN, dAvgAcc,  dAvgTrnTime, dAvgTstTime, dAvgMaxTrnTime);
					fprintf(fsptr, "%d:%d\t%.4f\t%.4f\t%.4f\t%.4f\n", param.NTOPICS, param.BURN_IN, sqrt(dStdAcc), sqrt(dStdTrnTime), sqrt(dStdTstTime), sqrt(dStdMaxTrnTime));
					fflush(fsptr);
				}
			}
			fclose(fsptr);
			delete[] acc;
			delete[] trnTime;
			delete[] tstTime;
			delete[] maxTrnTime;
		}
		else if ( strcmp(argv[1], "-multitask") == 0 ) 
		{
			double* acc				= new double[nRunTime];
			double* trnAcc			= new double[nRunTime];
			double* trnTime			= new double[nRunTime];
			double* tstTime			= new double[nRunTime];
			double* trnTimeEta		= new double[nRunTime];
			double* trnTimeLambda	= new double[nRunTime];
			FILE *fsptr = fopen("sensitivity_analysis_20ng_multi-task.txt", "a");
			fprintf(fsptr, "=== sensitivity to ell ===\n");
			for ( int i=1; i<10; i++ ) {
				//param.NTOPICS = (i+1) * 10;
				//param.INITIAL_ALPHA = 0.00001 * pow(10, 0.25*i); 
				//param.DELTA_ELL = pow((1 + i), 2.0);
				param.BURN_IN = i;
				for ( int m=0; m<nRunTime; m++ ) {
					if (param.INNER_CV_C) {
						Corpus* corpus = new Corpus();
						corpus->read_data_binary(param.train_filename);
						corpus->shuffle();
						param.INITIAL_C = inner_cv_C(corpus, &param);
						printf("\n\nBest C: %5.10f\n\n", param.INITIAL_C);
						delete corpus;
					}
					if (param.INNER_CV_ALPHA) {
						Corpus* corpus = new Corpus();
						corpus->read_data_binary(param.train_filename);
						corpus->shuffle();
						param.INITIAL_ALPHA = inner_cv_alpha(corpus, &param);
						printf("\n\nBest ALPHA: %5.10f\n\n", param.INITIAL_ALPHA);
						delete corpus;
					}

					MultiTaskGibbsMedLDA model(&param);
					model.init_est();
					model.train();
					model.save_model(param.model_filename);

					Corpus *test_corpus = new Corpus();
					// Read data and remove unseen words
					test_corpus->read_data_oov(param.test_filename, model.m_nV);
					model.init_inf(test_corpus);
					double tAvgEnt = 0;
					double dAcc = model.inference(&tAvgEnt);
					model.save_results(tAvgEnt);
					printf("Accuracy: %5.10f\n", dAcc);
					printf("testAverageEntropy: %5.10f\n", tAvgEnt);

					acc[m]				=  dAcc;
					trnAcc[m]			=  model.m_dTrainAcc;
					trnTime[m]			=  model.m_dTrainTime;
					trnTimeEta[m]		=  model.m_dTrainTime_Eta;
					trnTimeLambda[m]	=  model.m_dTrainTime_Lambda;
					tstTime[m]			=  model.m_dTestTime;
				}

				// save the average performance.
				double dAvgAcc = 0, dAvgTrnAcc = 0, dAvgTrnTime = 0, dAvgTstTime = 0, dAvgTrnTimeEta = 0, dAvgTrnTimeLambda = 0;
				double dStdAcc				= compute_mean_var( dAvgAcc, acc, nRunTime );
				double dStdTrnAcc			= compute_mean_var( dAvgTrnAcc, trnAcc, nRunTime);
				double dStdTrnTime			= compute_mean_var( dAvgTrnTime, trnTime, nRunTime );
				double dStdTrnTimeEta		= compute_mean_var( dAvgTrnTimeEta, trnTimeEta, nRunTime );
				double dStdTrnTimeLambda	= compute_mean_var( dAvgTrnTimeLambda, trnTimeLambda, nRunTime );
				double dStdTstTime			= compute_mean_var( dAvgTstTime, tstTime, nRunTime );
				printf("avg-acc: %.4f; %d\n", dAvgAcc, nRunTime);

				FILE *fptr = fopen(param.res_filename, "a");
				fprintf(fptr, "(k: %d, c: %.3f, NuSqInv: %.3f, alpha: %.3f, beta: %.3f, ell: %d, em-iter: %d, burn-in: %d, samples: %d) avgAcc: %.4f, %.4f; avgTrnTime: %.4f, %.4f; avgTstTime: %.4f, %.4f; trnAcc: %.4f, %.4f; eta: %.4f, lambda: %.4f\n", 
					param.NTOPICS, param.INITIAL_C, param.INITIAL_NUSQINV, param.INITIAL_ALPHA, param.INITIAL_BETA,
					(int)param.DELTA_ELL, param.EM_MAX_ITER, param.BURN_IN, param.MAX_SAMPLES,
					dAvgAcc, dStdAcc, dAvgTrnTime, dStdTrnTime, dAvgTstTime, dStdTstTime, 
					dAvgTrnAcc, dStdTrnAcc, dAvgTrnTimeEta, dAvgTrnTimeLambda );
				fflush(fptr);
				fclose(fptr);

				//fprintf(fsptr, "%d\t%.10f\t%.4f\t%.4f\t%.4f\t%.4f\t%.4f\t%.4f\n", param.NTOPICS, param.INITIAL_ALPHA, dAvgAcc, sqrt(dStdAcc), dAvgTrnTime, sqrt(dStdTrnTime), dAvgTstTime, sqrt(dStdTstTime));
				fprintf(fsptr, "(k: %d, c: %.3f, NuSqInv: %.3f, alpha: %.3f, beta: %.3f, ell: %d, burn-in: %d, samples: %d)\t%.4f\t%.4f\t%.4f\t%.4f\t%.4f\t%.4f\t%.4f\t%.4f\t%.4f\t%.4f\n", 
					param.NTOPICS, param.INITIAL_C, param.INITIAL_NUSQINV, param.INITIAL_ALPHA, param.INITIAL_BETA,
					(int)param.DELTA_ELL, param.BURN_IN, param.MAX_SAMPLES,
					dAvgAcc, dStdAcc, dAvgTrnTime, dStdTrnTime, dAvgTstTime, dStdTstTime, dAvgTrnAcc, dStdTrnAcc, dAvgTrnTimeEta, dAvgTrnTimeLambda );
				fflush(fsptr);
			}
			fclose(fsptr);
			delete[] acc;
			delete[] trnAcc;
			delete[] trnTime;
			delete[] trnTimeEta;
			delete[] trnTimeLambda;
			delete[] tstTime;
		}
		else if ( strcmp(argv[1], "-est") == 0 ) 
		{
			if (param.INNER_CV_C)
			{
				Corpus* corpus = new Corpus();
				corpus->read_data_binary(param.train_filename);
				corpus->shuffle();
				param.INITIAL_C = inner_cv_C(corpus, &param);
				printf("\n\nBest C: %5.10f\n\n", param.INITIAL_C);
				delete corpus;
			}

			if (param.INNER_CV_ALPHA)
			{
				Corpus* corpus = new Corpus();
				corpus->read_data_binary(param.train_filename);
				corpus->shuffle();
				param.INITIAL_ALPHA = inner_cv_alpha(corpus, &param);
				printf("\n\nBest ALPHA: %5.10f\n\n", param.INITIAL_ALPHA);
				delete corpus;
			}

			GibbsMedLDA model(&param);
			model.init_est();
			model.run_em();
			model.save_model(param.model_filename);
		}
		else if ( strcmp(argv[1], "-estc") == 0 ) 
		{
			GibbsMedLDA model(&param);
			model.init_estc(argv[7]);
			model.run_em();
			model.save_model(param.model_filename);
		}
		else if ( strcmp(argv[1], "-inf") == 0 ) 
		{
			GibbsMedLDA model(&param);
			printf("Model file: %s\n", param.model_filename);
			model.init_inf(param.model_filename);
			double tAvgEnt = 0;
			double dAcc = model.inference(&tAvgEnt);
			model.save_results(tAvgEnt);
			printf("Accuracy: %5.10f\n", dAcc);
		}
		else
		{
			printf("Usages: medlda -estinf [K] [lables] [Alpha] [SampleLag] [C] [MAX-EM-ITER] [folds] [training_data] [test_data]\n");
			//printf("        medlda -est [K] [lables] [Alpha] [SampleLag] [C] [folds] [training_data]\n");
			//printf("        medlda -estc [K] [lables] [Alpha] [SampleLag] [C] [folds] [prev_state_file]\n");
			//printf("        medlda -inf [K] [lables] [Alpha] [SampleLag] [C] [folds] [test_data]\n");
		}
	}
	else
	{
		//printf("Usages: medlda -estinf [K] [lables] [Alpha] [SampleLag] [C] [folds] [training_data] [test_data]\n");
		printf("Usages: medlda -estinf [K] [lables] [Alpha] [SampleLag] [C] [MAX-EM-ITER] [folds] [training_data] [test_data]\n");
		//printf("        medlda -est [K] [lables] [Alpha] [SampleLag] [C] [folds] [training_data]\n");
		//printf("        medlda -estc [K] [lables] [Alpha] [SampleLag] [C] [folds] [prev_state_file]\n");
		//printf("        medlda -inf [K] [lables] [Alpha] [SampleLag] [C] [folds] [test_data]\n");
	}
	return 0;
}

double inner_cv_C(Corpus* corpus,  Params* param)
{
	Params* inner_param = param->copy();
	sprintf(inner_param->dir, "%s/innerCVC", param->dir);
	//make_directory(inner_param->dir);
	mkdir(inner_param->dir, 0777);

	double dBestC = 0.0;
	double dBestAccuracy = 0.0;
	double dAcc = 0;
	double tAvgEnt = 0;
	for( int k = 0; k < param->CV_PARAMNUM_C; ++k)
	{
		printf("\n\n$$$ Learning with C: %.4f $$$ \n\n", param->cv_param_vec_C[k]);
		inner_param->INITIAL_C = param->cv_param_vec_C[k];
		double dAccuracy = 0;
		for (int i = 0; i < param->INNER_FOLDNUM_C; ++i)
		{
			printf("\n FOLD: %d \n\n", i);
			Corpus* trn_corpus = corpus->get_traindata(inner_param->INNER_FOLDNUM_C, i); 	
			Corpus* tst_corpus = corpus->get_testdata(inner_param->INNER_FOLDNUM_C, i); 	

			GibbsMedLDA* gml = new GibbsMedLDA(inner_param);
			gml->init_est(trn_corpus);
			gml->run_em();
			gml->init_inf(tst_corpus);
			dAcc = gml->inference(&tAvgEnt);
			dAccuracy += dAcc / param->INNER_FOLDNUM_C;
			delete gml;
		}
		printf("@@@ Avg Accuracy: %.4f\n\n", dAccuracy);
		if (dAccuracy > dBestAccuracy)
		{
			dBestAccuracy = dAccuracy;
			dBestC		  = inner_param->INITIAL_C;
		}
	}
	delete inner_param;
	return dBestC;
}

double inner_cv_alpha(Corpus* corpus, Params* param)
{
	char buf[512];
	Params* inner_param = param->copy();
	sprintf(inner_param->dir, "%s/innerCVAlpha", param->dir);
	sprintf(inner_param->model_filename, "%s/model-final.txt", inner_param->dir);
	//make_directory(inner_param->dir);
	mkdir(inner_param->dir, 0777);

	double dBestALPHA 	= 0.0;
	double dBestAccuracy = 0.0;
	double dAcc 			= 0;
	double tAvgEnt = 0;
	for( int k = 0; k < param->CV_PARAMNUM_ALPHA; ++k)
	{
		printf("\n\n$$$ Learning with ALPHA: %.4f $$$ \n\n", param->cv_param_vec_ALPHA[k]);
		inner_param->INITIAL_ALPHA = param->cv_param_vec_ALPHA[k];
		double dAccuracy = 0;
		for (int i = 0; i < param->INNER_FOLDNUM_ALPHA; ++i)
		{
			printf("\n FOLD: %d \n\n", i);
			Corpus* trn_corpus = corpus->get_traindata(inner_param->INNER_FOLDNUM_ALPHA, i); 	
			Corpus* tst_corpus = corpus->get_testdata(inner_param->INNER_FOLDNUM_ALPHA, i); 	

			GibbsMedLDA* gml = new GibbsMedLDA(inner_param);
			gml->init_est(trn_corpus);
			gml->run_em();
			gml->init_inf(tst_corpus);
			dAcc = gml->inference(&tAvgEnt);
			printf("[K: %d; Alpha: %f; Fold: %d] dAcc: %.5f\n", inner_param->NTOPICS, inner_param->INITIAL_ALPHA, i, dAcc);
			dAccuracy += dAcc / param->INNER_FOLDNUM_ALPHA;
			delete gml;
		}
		printf("@@@ [K: %d; Alpha: %f] Avg_Acc: %.5f\n\n", inner_param->NTOPICS, inner_param->INITIAL_ALPHA, dAccuracy);
		if (dAccuracy > dBestAccuracy)
		{
			dBestAccuracy = dAccuracy;
			dBestALPHA	  = inner_param->INITIAL_ALPHA;
		}
	}
	delete inner_param;
	return dBestALPHA;
}

double save_results_multi(double **predScore, int *gndLabel, int nTstData, double dTrainTime, double dTestTime, Params *param)
{
	// winner-take-all
	double dAcc = 0;
	int *predLabel = new int[nTstData];
	for ( int d=0; d<nTstData; d++ ) {
		predLabel[d] = -1;
		double dMaxScore = 0;
		for ( int m=0; m<param->NLABELS; m++ ) {
			if ( predLabel[d] == -1 || predScore[m][d] > dMaxScore ) {
				predLabel[d] = m;
				dMaxScore = predScore[m][d];
			}
		}
		if ( predLabel[d] == gndLabel[d] ) dAcc ++;
	}
	dAcc /= nTstData;

	char buf[512];
	sprintf(buf, "%s/evl-performance.txt", param->dir);
	FILE* fileptr = fopen(buf, "w");
	fprintf(fileptr, "num docs: %d\n", nTstData);
	fprintf(fileptr, "num topics: %d\n", param->NTOPICS );
	fprintf(fileptr, "cost parameter l: %d\n", (int)param->DELTA_ELL);
	fprintf(fileptr, "Alpha: %5.10f\n", param->INITIAL_ALPHA ); 
	fprintf(fileptr, "C:	%5.10f\n\n", param->INITIAL_C);
	fprintf(fileptr, "Accuracy: %5.10f\n", dAcc);
	fclose(fileptr);

	sprintf(buf, "%s/pred_results.txt", param->dir);
	fileptr = fopen(buf, "w");
	for ( int d=0; d<nTstData; d++ ) {
		fprintf(fileptr, "%d\t%d\t", gndLabel[d], predLabel[d]);
		for ( int m=0; m<param->NLABELS; m++ ) {
			fprintf(fileptr, "%.3f ", predScore[m][d]);
		}
		fprintf(fileptr, "\n");
	}
	fclose(fileptr);

	// overall-results
	FILE *fres = fopen(param->res_filename, "a");
	fprintf(fres, "(k: %d, c: %.3f, alpha: %.3f, beta: %.3f, ell: %d, em-iter: %d, burn-in: %d, samples: %d) acc: %.4f, train-time: %.4f, test-time: %.4f\n", 
		param->NTOPICS, param->INITIAL_C, param->INITIAL_ALPHA, param->INITIAL_BETA,
		(int)param->DELTA_ELL, param->EM_MAX_ITER, param->BURN_IN, param->MAX_SAMPLES,
		dAcc, dTrainTime, dTestTime);
	fflush(fres);
	fclose(fres);

	printf(">>> Saving results & evaluating performace finished! \n");
	delete[] predLabel;

	return dAcc;
}
