// (C) Copyright 2009, Jun Zhu (junzhu [at] cs [dot] cmu [dot] edu)

// This file is part of MedLDA.

// MedLDA is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your
// option) any later version.

// MedLDA is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
// USA

#include "params.h"
#include <string>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
using namespace std;

Params::Params(void)
{
	train_filename 		= new char[512];
	test_filename 			= new char[512];
	mapper_filename		= new char[512];
	model_filename			= new char[512];
	dir						= new char[512];
	res_filename            = new char[512];
	cv_param_vec_C			= NULL;
	cv_param_vec_ALPHA	= NULL;
	DELTA_ELL = 16;
	INITIAL_NUSQINV = 1;
}

Params::~Params(void)
{
	delete[] train_filename;
	delete[] test_filename;
	delete[] mapper_filename;
	delete[] model_filename;
	delete[] dir;
	delete[] res_filename;
	if(cv_param_vec_C) delete[] cv_param_vec_C;
	if(cv_param_vec_ALPHA) delete[] cv_param_vec_ALPHA;
}

void Params::read_settings(char* filename)
{
  int ret = 0;
	INITIAL_C = 1;
	INITIAL_NUSQINV = 1;
	FILE* fileptr;
	char buf[100];
	fileptr = fopen(filename, "r");

	// parameters for the strategy that sets stopping criterion for burn-in stage.
	fscanf(fileptr, "em max iter %d\n", &EM_MAX_ITER);
	fscanf(fileptr, "em convergence %lf\n", &EM_CONVERGED);
	fscanf(fileptr, "gibbs max iter %d\n", &NITERS);
	fscanf(fileptr, "gibbs convergence %lf\n", &GIBBS_CONVERGED); 
	fscanf(fileptr, "sampling lag %d\n", &SAMPLING_LAG); 

	// parameters for the standard burn-in v.s draw samples strategy
	fscanf(fileptr, "gibbs burn-in %d\n", &BURN_IN);
	fscanf(fileptr, "gibbs max-samples %d\n", &MAX_SAMPLES);

	// hyper-parameters
	fscanf(fileptr, "init alpha %lf\n", &INITIAL_ALPHA);
	fscanf(fileptr, "init beta %lf\n", &INITIAL_BETA);
	fscanf(fileptr, "delta_ell %lf\n", &DELTA_ELL);
	fscanf(fileptr, "sample_save_step %d\n", &ITER_SAVE_STEP);
	fscanf(fileptr, "save_top_words %d\n", &SAVE_TOP_WORDS);

	fscanf(fileptr, "inner-cv-C %s\n", buf);
	INNER_CV_C = (strcmp(buf, "true") == 0) ? true : false;
	fscanf(fileptr, "inner_foldnum_C %d\n", &INNER_FOLDNUM_C);
	fscanf(fileptr, "cv_paramnum_C %d\n", &CV_PARAMNUM_C);
	cv_param_vec_C = new double[CV_PARAMNUM_C];
	for (int i = 0 ; i < CV_PARAMNUM_C; ++i)
		fscanf(fileptr, "%lf\n", &cv_param_vec_C[i]);

	fscanf(fileptr, "inner-cv-alpha %s\n", buf);
	INNER_CV_ALPHA = (strcmp(buf, "true") == 0) ? true : false;
	fscanf(fileptr, "inner_foldnum_alpha %d\n", &INNER_FOLDNUM_ALPHA);
	fscanf(fileptr, "cv_paramnum_alpha %d\n", &CV_PARAMNUM_ALPHA);
	cv_param_vec_ALPHA = new double[CV_PARAMNUM_ALPHA];
	for (int i = 0 ; i < CV_PARAMNUM_ALPHA; ++i)
		fscanf(fileptr, "%lf\n", &cv_param_vec_ALPHA[i]);

	fscanf(fileptr, "mapper_filename: %s\n", mapper_filename); 
	fscanf(fileptr, "train_filename: %s\n", train_filename);
	fscanf(fileptr, "test_filename: %s\n", test_filename);
	fscanf(fileptr, "res_filename: %s\n", res_filename);
	fclose(fileptr);
}

Params* Params::copy()
{
	Params* pa = new Params();

	pa->EM_MAX_ITER 			= EM_MAX_ITER;
	pa->EM_CONVERGED			= EM_CONVERGED;
	pa->NITERS					= NITERS;
	pa->GIBBS_CONVERGED 		= GIBBS_CONVERGED;
	pa->SAMPLING_LAG			= SAMPLING_LAG;

	pa->BURN_IN                 = BURN_IN;
	pa->MAX_SAMPLES             = MAX_SAMPLES;

	pa->INITIAL_C				= INITIAL_C;
	pa->INITIAL_ALPHA			= INITIAL_ALPHA;
	pa->INITIAL_BETA			= INITIAL_BETA;

	pa->ITER_SAVE_STEP 		= ITER_SAVE_STEP;
	pa->SAVE_TOP_WORDS		= SAVE_TOP_WORDS;

	pa->NTOPICS					= NTOPICS;
	pa->NLABELS					= NLABELS;

	pa->NFOLDS					= NFOLDS;

	pa->INNER_CV_C				= INNER_CV_C;
	pa->INNER_FOLDNUM_C		= INNER_FOLDNUM_C;
	pa->CV_PARAMNUM_C			= CV_PARAMNUM_C;
	pa->cv_param_vec_C		= new double[pa->CV_PARAMNUM_C];
	memcpy(pa->cv_param_vec_C, cv_param_vec_C, sizeof(double) * pa->CV_PARAMNUM_C);

	pa->INNER_CV_ALPHA		= INNER_CV_ALPHA;
	pa->INNER_FOLDNUM_ALPHA	= INNER_FOLDNUM_ALPHA;
	pa->CV_PARAMNUM_ALPHA	= CV_PARAMNUM_ALPHA;
	pa->cv_param_vec_ALPHA	= new double[pa->CV_PARAMNUM_ALPHA];
	memcpy(pa->cv_param_vec_ALPHA, cv_param_vec_ALPHA, sizeof(double) * pa->CV_PARAMNUM_ALPHA);

	memcpy(pa->dir, dir, sizeof(char)* 512);
	memcpy(pa->train_filename, train_filename, sizeof(char)* 512);
	memcpy(pa->test_filename, test_filename, sizeof(char)* 512);
	memcpy(pa->mapper_filename, mapper_filename, sizeof(char)* 512);
	memcpy(pa->model_filename, model_filename, sizeof(char)*512);
	memcpy(pa->res_filename, res_filename, sizeof(char)*512);

	return pa;
}
