#ifndef _STATS_H
#define _STATS_H

#include "defs.h"
#include "params.h"
#include "corpus.h"
#include <stdlib.h>
#include <stdio.h>
#include <iostream>
#include <string.h>
#include <time.h>

using namespace std;

class TopK
{
public:
	TopK(int len): length(len), key(NULL), value(NULL)
	{
		key = new int[length];
		value = new double[length];
	}
	~TopK()
	{
		if (key) delete[] key;
		if (value) delete[] value;
	}
	void reset();
	void push_back(int id, double v);
	void sort();

private:
	void shiftdown(int pos, int rg);
	inline void exchange(int p1, int p2);

public:
	int length;
	int* key;
	double* value;
};

class Stats
{
public:
	Stats();
	~Stats();

	void new_stats(Corpus *cps, Params* param);
	void load(const char *filename, Params* param);
	void load(FILE* file, Params* param);
	void free();

	void save_wordassigned(const char *filename);
	void save_wordassigned(FILE *file);
	void save_topwords(const char* filename, Params* param);
	void save_label_only(const char* filename);

public:
	int nD; // # docs
	int nV; // # terms
	int nK; // # topics

	int **z;	// topic assignments for D X doc.lenght
	int **nw; // m_nw[i][j]: number of instances of word/term i assigned to topic j, size K x V
	int *nwsum; // m_nwsum[k]: # words assigned to the topic k, 1 X K.
	int **nd; // m_nd[i][j]: number of words in document i assigned to topic j, size m_D x K
	
	Corpus* corpus;
};
#endif

